@@ -1,8 +1,11 @@
-# -*- perl -*-
+#/usr/bin/perl
 
 package Bundle::DBD::CSV;
 
-$VERSION = "1.05";
+use strict;
+use warnings;
+
+our $VERSION = "1.14";
 
 1;
 
@@ -18,15 +21,15 @@ Bundle::DBD::CSV - A bundle to install the DBD::CSV driver
 
 =head1 CONTENTS
 
-DBI 1.614
+DBI 1.641
 
-Text::CSV_XS 0.73
+Text::CSV_XS 1.37
 
-SQL::Statement 1.31
+SQL::Statement 1.412
 
-DBD::File 0.40
+DBD::File 0.44
 
-DBD::CSV 0.31
+DBD::CSV 0.54
 
 =head1 DESCRIPTION
 
@@ -41,7 +44,7 @@ This module is currently maintained by
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2009-2010 by H.Merijn Brand
+Copyright (C) 2009-2018 by H.Merijn Brand
 Copyright (C) 2004-2009 by Jeff Zucker
 Copyright (C) 1998-2004 by Jochen Wiedmann
 
