@@ -1,4 +1,4 @@
-#!/pro/bin/perl
+#!/usr/bin/perl
 #
 #   DBD::CSV - A DBI driver for CSV and similar structured files
 #
@@ -6,21 +6,10 @@
 #
 #	H.Merijn Brand <h.m.brand@xs4all.nl>
 #
-#   The original author is Jochen Wiedmann.
-#   Then maintained by Jeff Zucker
-#
-#   Copyright (C) 2010 by H.Merijn Brand
-#   Copyright (C) 2004 by Jeff Zucker
-#   Copyright (C) 1998 by Jochen Wiedmann
-#
-#   All rights reserved.
-#
-#   You may distribute this module under the terms of either the GNU
-#   General Public License or the Artistic License, as specified in
-#   the Perl README file.
+#   See for full acknowledgements the last two pod sections in this file
 
-require 5.005003;
 use strict;
+use warnings;
 
 require DynaLoader;
 require DBD::File;
@@ -30,19 +19,19 @@ package DBD::CSV;
 
 use strict;
 
-use vars qw( @ISA $VERSION $drh $err $errstr $sqlstate );
+use vars qw( @ISA $VERSION $ATTRIBUTION $drh $err $errstr $sqlstate );
 
 @ISA =   qw( DBD::File );
 
-$VERSION  = "0.31";
+$VERSION     = "0.54";
+$ATTRIBUTION = "DBD::CSV $DBD::CSV::VERSION by H.Merijn Brand";
 
 $err      = 0;		# holds error code   for DBI::err
 $errstr   = "";		# holds error string for DBI::errstr
 $sqlstate = "";         # holds error state  for DBI::state
 $drh      = undef;	# holds driver handle once initialized
 
-sub CLONE		# empty method: prevent warnings when threads are cloned
-{
+sub CLONE {		# empty method: prevent warnings when threads are cloned
     } # CLONE
 
 # --- DRIVER -------------------------------------------------------------------
@@ -72,19 +61,15 @@ use vars qw( @ISA @CSV_TYPES );
     Text::CSV_XS::NV (), # SQL_DOUBLE
     );
 
-@DBD::CSV::dr::ISA = qw( DBD::File::dr );
+our @ISA = qw( DBD::File::dr );
 
-$DBD::CSV::dr::imp_data_size     = 0;
-$DBD::CSV::dr::data_sources_attr = undef;
+our $imp_data_size     = 0;
+our $data_sources_attr = undef;
 
-$DBD::CSV::ATTRIBUTION = "DBD::CSV $DBD::CSV::VERSION by H.Merijn Brand";
-
-sub connect
-{
+sub connect {
     my ($drh, $dbname, $user, $auth, $attr) = @_;
     my $dbh = $drh->DBD::File::dr::connect ($dbname, $user, $auth, $attr);
-    $dbh->{f_meta} ||= {};
-    $dbh->{Active}   = 1;
+    $dbh and $dbh->{Active} = 1;
     $dbh;
     } # connect
 
@@ -94,53 +79,46 @@ package DBD::CSV::db;
 
 use strict;
 
-$DBD::CSV::db::imp_data_size = 0;
+our $imp_data_size = 0;
+our @ISA = qw( DBD::File::db );
 
-@DBD::CSV::db::ISA = qw( DBD::File::db );
-
-sub set_versions
-{
+sub set_versions {
     my $this = shift;
     $this->{csv_version} = $DBD::CSV::VERSION;
     return $this->SUPER::set_versions ();
     } # set_versions
 
-if ($DBD::File::VERSION <= 0.38) {
-    # Map csv_tables to f_meta.
-    # Not absolutely needed, but otherwise I have to write two test suites
-    *STORE = sub {
-	my ($self, @attr) = @_;
-	@attr && $attr[0] eq "csv_tables" and $attr[0] = "f_meta";
-	$self->SUPER::STORE (@attr);
-	}; # STORE
-
-    *FETCH = sub {
-	my ($self, @attr) = @_;
-	@attr && $attr[0] eq "csv_tables" and $attr[0] = "f_meta";
-	$self->SUPER::FETCH (@attr);
-	}; # FETCH
-
-    *DBI::db::csv_versions = *csv_versions = sub {
-	join "\n",
-	    "DBD::CSV     $DBD::CSV::VERSION using Text::CSV_XS-$Text::CSV_XS::VERSION",
-	    "  DBD::File  $DBD::File::VERSION",
-	    "DBI          $DBI::VERSION",
-	    "OS           $^O",
-	    "Perl         $]";
-	}; # csv_versions
-    }
-
 my %csv_xs_attr;
 
-sub init_valid_attributes
-{
+sub init_valid_attributes {
     my $dbh = shift;
 
+    # Straight from Text::CSV_XS.pm
     my @xs_attr = qw(
-	allow_loose_escapes allow_loose_quotes allow_whitespace
-	always_quote auto_diag binary blank_is_undef empty_is_undef
-	eol escape_char keep_meta_info quote_char quote_null
-	quote_space sep_char types verbatim );
+	eol
+	sep_char
+	quote_char
+	escape_char
+	binary
+	decode_utf8
+	auto_diag
+	diag_verbose
+	blank_is_undef
+	empty_is_undef
+	allow_whitespace
+	allow_loose_quotes
+	allow_loose_escapes
+	allow_unquoted_escape
+	always_quote
+	quote_empty
+	quote_space
+	escape_null
+	quote_binary
+	keep_meta_info
+	verbatim
+	types
+	callbacks
+	);
     @csv_xs_attr{@xs_attr} = ();
 
     $dbh->{csv_xs_valid_attrs} = [ @xs_attr ];
@@ -149,7 +127,7 @@ sub init_valid_attributes
 
 	class tables in csv_in out csv_out skip_first_row
 
-	null sep quote escape
+	null sep quote escape bom
 	)};
 
     $dbh->{csv_readonly_attrs} = { };
@@ -159,8 +137,7 @@ sub init_valid_attributes
     return $dbh->SUPER::init_valid_attributes ();
     } # init_valid_attributes
 
-sub get_csv_versions
-{
+sub get_csv_versions {
     my ($dbh, $table) = @_;
     $table ||= "";
     my $class = $dbh->{ImplementorClass};
@@ -175,7 +152,21 @@ sub get_csv_versions
     my $dtype = $meta->{csv_class};
     $dvsn and $dtype .= " ($dvsn)";
     return sprintf "%s using %s", $dbh->{csv_version}, $dtype;
-    } # get_csv_versions 
+    } # get_csv_versions
+
+sub get_info {
+    my ($dbh, $info_type) = @_;
+    require  DBD::CSV::GetInfo;
+    my $v = $DBD::CSV::GetInfo::info{int ($info_type)};
+    ref $v eq "CODE" and $v = $v->($dbh);
+    return $v;
+    } # get_info
+
+sub type_info_all {
+#   my $dbh = shift;
+    require   DBD::CSV::TypeInfo;
+    return [@$DBD::CSV::TypeInfo::type_info_all];
+    } # type_info_all
 
 # --- STATEMENT ----------------------------------------------------------------
 
@@ -183,161 +174,41 @@ package DBD::CSV::st;
 
 use strict;
 
-$DBD::CSV::st::imp_data_size = 0;
-
-@DBD::CSV::st::ISA = qw(DBD::File::st);
-
-$DBD::File::VERSION <= 0.38 and *FETCH = sub {
-    my ($sth, $attr) = @_;
-
-    my ($struct, @coldefs, @colnames);
-
-    # Being a bit dirty here, as SQL::Statement::Structure does not offer
-    # me an interface to the data I want
-    $struct   = $sth->{f_stmt}{struct} || {};
-    @coldefs  = @{ $struct->{column_defs} || [] };
-    @colnames = map { $_->{name} || $_->{value} } @coldefs;
-
-    # dangerous: this accesses the table_defs information from last CREATE TABLE statement
-    $attr eq "TYPE"      and
-	return [ map { $struct->{table_defs}{columns}{$_}{data_type}   || "CHAR" }
-		    @colnames ];
-
-    $attr eq "PRECISION" and
-	return [ map { $struct->{table_defs}{columns}{$_}{data_length} || 0 }
-		    @colnames ];
-
-    $attr eq "NULLABLE"  and
-	return [ map { ( grep m/^NOT NULL$/ =>
-		    @{ $struct->{table_defs}{columns}{$_}{constraints} || [] } )
-		       ? 0 : 1 }
-		    @colnames ];
-
-    return $sth->SUPER::FETCH ($attr);
-    }; # FETCH
+our $imp_data_size = 0;
+our @ISA = qw( DBD::File::st );
 
 package DBD::CSV::Statement;
 
 use strict;
-use DBD::File;
 use Carp;
 
-@DBD::CSV::Statement::ISA = qw(DBD::File::Statement);
-
-# open_table (0 is used up to and including DBI-1.1611
-# Later versions use open_file (see DBD::CSV::Table)
-
-$DBD::File::VERSION <= 0.38 and *open_table = sub {
-    my ($self, $data, $table, $createMode, $lockMode) = @_;
-
-    my $dbh    = $data->{Database};
-    my $tables = $dbh->{f_meta};
-       $tables->{$table} ||= {};
-    my $meta   = $tables->{$table} || {};
-    my $csv_in = $meta->{csv_in} || $dbh->{csv_csv_in};
-    unless ($csv_in) {
-	my %opts  = ( binary => 1, auto_diag => 1 );
-
-	# Allow specific Text::CSV_XS options
-	foreach my $key (grep m/^csv_/ => keys %$dbh) {
-	    (my $attr = $key) =~ s/csv_//;
-	    $attr =~ m{^(?: eol | sep | quote | escape	# Handled below
-			  | tables | sql_parser_object	# Not for Text::CSV_XS
-			  | sponge_driver | version	# internal
-			  )$}x and next;
-	    $opts{$attr} = $dbh->{$key};
-	    }
-	delete $opts{null} and
-	    $opts{blank_is_undef} = $opts{always_quote} = 1;
-
-	my $class = $meta->{class} || $dbh->{csv_class} || "Text::CSV_XS";
-	my $eol   = $meta->{eol}   || $dbh->{csv_eol}   || "\r\n";
-	$eol =~ m/^\A(?:[\r\n]|\r\n)\Z/ or $opts{eol} = $eol;
-	for ([ "sep",    ',' ],
-	     [ "quote",  '"' ],
-	     [ "escape", '"' ],
-	     ) {
-	    my ($attr, $def) = ($_->[0]."_char", $_->[1]);
-	    $opts{$attr} =
-		exists $meta->{$attr} ? $meta->{$attr} :
-		    exists $dbh->{"csv_$attr"} ? $dbh->{"csv_$attr"} : $def;
-	    }
-	$meta->{csv_in}  = $class->new (\%opts) or
-	    $class->error_diag;
-	$opts{eol} = $eol;
-	$meta->{csv_out} = $class->new (\%opts) or
-	    $class->error_diag;
-	}
-    my $file = $meta->{file} || $table;
-    my $tbl  = $self->SUPER::open_table ($data, $file, $createMode, $lockMode);
-    if ($tbl && $tbl->{fh}) {
-	$tbl->{csv_csv_in}  = $meta->{csv_in};
-	$tbl->{csv_csv_out} = $meta->{csv_out};
-	if (my $types = $meta->{types}) {
-	    # The 'types' array contains DBI types, but we need types
-	    # suitable for Text::CSV_XS.
-	    my $t = [];
-	    for (@{$types}) {
-		$_ = $_
-		    ? $DBD::CSV::dr::CSV_TYPES[$_ + 6] || Text::CSV_XS::PV ()
-		    : Text::CSV_XS::PV ();
-		push @$t, $_;
-		}
-	    $tbl->{types} = $t;
-	    }
-	if ( !$createMode and
-	     !$self->{ignore_missing_table} and $self->{command} ne "DROP") {
-	    my $array;
-	    my $skipRows = exists $meta->{skip_rows}
-		? $meta->{skip_rows}
-		: exists $meta->{col_names} ? 0 : 1;
-	    if ($skipRows--) {
-		$array = $tbl->fetch_row ($data) or croak "Missing first row";
-		unless ($self->{raw_header}) {
-		    s/\W/_/g for @$array;
-		    }
-		$tbl->{col_names} = $array;
-		while ($skipRows--) {
-		    $tbl->fetch_row ($data);
-		    }
-		}
-	    $tbl->{first_row_pos} = $tbl->{fh}->tell ();
-	    exists $meta->{col_names} and
-		$array = $tbl->{col_names} = $meta->{col_names};
-	    if (!$tbl->{col_names} || !@{$tbl->{col_names}}) {
-		# No column names given; fetch first row and create default
-		# names.
-		my $ar = $tbl->{cached_row} = $tbl->fetch_row ($data);
-		$array = $tbl->{col_names};
-		push @$array, map { "col$_" } 0 .. $#$ar;
-		}
-	    my $i = 0;
-	    $tbl->{col_nums}{$_} = $i++ for @$array;
-	    }
-	}
-    $tbl;
-    }; # open_table
+our @ISA = qw( DBD::File::Statement );
 
 package DBD::CSV::Table;
 
 use strict;
-use DBD::File;
 use Carp;
 
-@DBD::CSV::Table::ISA = qw(DBD::File::Table);
+our @ISA = qw( DBD::File::Table );
 
-sub bootstrap_table_meta
-{
+#sub DESTROY {
+#    my $self = shift or return;
+#
+#    $self->{meta} and delete $self->{meta}{csv_in};
+#    } # DESTROY
+
+sub bootstrap_table_meta {
     my ($self, $dbh, $meta, $table) = @_;
     $meta->{csv_class} ||= $dbh->{csv_class} || "Text::CSV_XS";
     $meta->{csv_eol}   ||= $dbh->{csv_eol}   || "\r\n";
     exists $meta->{csv_skip_first_row} or
 	$meta->{csv_skip_first_row} = $dbh->{csv_skip_first_row};
+    exists $meta->{csv_bom} or
+	$meta->{csv_bom} = exists $dbh->{bom} ? $dbh->{bom} : $dbh->{csv_bom};
     $self->SUPER::bootstrap_table_meta ($dbh, $meta, $table);
     } # bootstrap_table_meta
 
-sub init_table_meta
-{
+sub init_table_meta {
     my ($self, $dbh, $meta, $table) = @_;
 
     $self->SUPER::init_table_meta ($dbh, $table, $meta);
@@ -352,7 +223,8 @@ sub init_table_meta
 	    exists $dbh->{"csv_$attr"} and $opts{$attr} = $dbh->{"csv_$attr"};
 	    }
 	$dbh->{csv_null} || $meta->{csv_null} and
-	    $opts{blank_is_undef} = $opts{always_quote} = 1;
+	    $opts{Text::CSV_XS->version < 1.18 ? "always_quote" : "quote_empty"} =
+	    $opts{blank_is_undef} = 1;
 
 	my $class = $meta->{csv_class};
 	my $eol   = $meta->{csv_eol};
@@ -379,8 +251,7 @@ my %compat_map = map { $_ => "csv_$_" }
 
 __PACKAGE__->register_compat_map (\%compat_map);
 
-sub table_meta_attr_changed
-{
+sub table_meta_attr_changed {
     my ($class, $meta, $attr, $value) = @_;
 
     (my $csv_attr = $attr) =~ s/^csv_//;
@@ -394,12 +265,11 @@ sub table_meta_attr_changed
     $class->SUPER::table_meta_attr_changed ($meta, $attr, $value);
     } # table_meta_attr_changed
 
-$DBD::File::VERSION > 0.38 and *open_file = sub {
+sub open_data {
     my ($self, $meta, $attrs, $flags) = @_;
     $self->SUPER::open_file ($meta, $attrs, $flags);
 
-    my $tbl = $meta;
-    if ($tbl && $tbl->{fh}) {
+    if ($meta && $meta->{fh}) {
 	$attrs->{csv_csv_in}  = $meta->{csv_in};
 	$attrs->{csv_csv_out} = $meta->{csv_out};
 	if (my $types = $meta->{types}) {
@@ -413,7 +283,7 @@ $DBD::File::VERSION > 0.38 and *open_file = sub {
 		    : Text::CSV_XS::PV ();
 		push @$t, $_;
 		}
-	    $tbl->{types} = $t;
+	    $meta->{types} = $t;
 	    }
 	if (!$flags->{createMode}) {
 	    my $array;
@@ -424,75 +294,99 @@ $DBD::File::VERSION > 0.38 and *open_file = sub {
 		    : exists $meta->{col_names} ? 0 : 1;
 	    defined $meta->{skip_rows} or
 		$meta->{skip_rows} = $skipRows;
+	    if ($meta->{csv_bom}) {
+		my @hdr = $attrs->{csv_csv_in}->header ($meta->{fh}) or
+		    croak "Failed using the header row: ".$attrs->{csv_csv_in}->error_diag;
+		$meta->{col_names} ||= \@hdr;
+		$skipRows and $skipRows = 0;
+		}
 	    if ($skipRows--) {
-		$array = $attrs->{csv_csv_in}->getline ($tbl->{fh}) or
+		$array = $attrs->{csv_csv_in}->getline ($meta->{fh}) or
 		    croak "Missing first row due to ".$attrs->{csv_csv_in}->error_diag;
 		unless ($meta->{raw_header}) {
 		    s/\W/_/g for @$array;
 		    }
-		$tbl->{col_names} = $array;
+		defined $meta->{col_names} or
+		    $meta->{col_names} = $array;
 		while ($skipRows--) {
-		    $tbl->{csv_csv_in}->getline ($tbl->{fh});
+		    $attrs->{csv_csv_in}->getline ($meta->{fh});
 		    }
 		}
-	    $tbl->{first_row_pos} = $tbl->{fh}->tell ();
+	    # lockMode is set 1 for DELETE, INSERT or UPDATE
+	    # no other case need seeking
+	    $flags->{lockMode} and # $meta->{fh}->can ("tell") and
+		$meta->{first_row_pos} = $meta->{fh}->tell ();
 	    exists $meta->{col_names} and
-		$array = $tbl->{col_names} = $meta->{col_names};
-	    if (!$tbl->{col_names} || !@{$tbl->{col_names}}) {
+		$array = $meta->{col_names};
+	    if (!$meta->{col_names} || !@{$meta->{col_names}}) {
 		# No column names given; fetch first row and create default
 		# names.
-		my $ar = $tbl->{cached_row} =
-		    $tbl->{csv_csv_in}->getline ($tbl->{fh});
-		$array = $tbl->{col_names};
+		my $ar = $meta->{cached_row} =
+		    $attrs->{csv_csv_in}->getline ($meta->{fh});
+		$array = $meta->{col_names};
 		push @$array, map { "col$_" } 0 .. $#$ar;
 		}
-	    my $i = 0;
-	    $tbl->{col_nums}{$_} = $i++ for @$array; # XXX not necessary for DBI > 1.611
 	    }
 	}
-    }; # open_file
+    } # open_file
 
-sub fetch_row
-{
+no warnings 'once';
+$DBI::VERSION < 1.623 and
+    *open_file = \&open_data;
+use warnings;
+
+sub _csv_diag {
+    my @diag = $_[0]->error_diag;
+    for (2, 3) {
+	defined $diag[$_] or $diag[$_] = "?";
+	}
+    return @diag;
+    } # _csv_diag
+
+sub fetch_row {
     my ($self, $data) = @_;
 
-    exists $self->{cached_row} and
-	return $self->{row} = delete $self->{cached_row};
+    my $tbl = $self->{meta};
 
-    my $tbl = $DBD::File::VERSION <= 0.38 ? $self : $self->{meta};
+    exists $tbl->{cached_row} and
+	return $self->{row} = delete $tbl->{cached_row};
 
     my $csv = $self->{csv_csv_in} or
 	return do { $data->set_err ($DBI::stderr, "Fetch from undefined handle"); undef };
 
-    my $fields;
-    eval { $fields = $csv->getline ($tbl->{fh}) };
+    my $fields = eval { $csv->getline ($tbl->{fh}) };
     unless ($fields) {
 	$csv->eof and return;
 
-	my @diag = $csv->error_diag;
-	my $file = $DBD::File::VERSION <= 0.38 ? $self->{file} : $tbl->{f_fqfn};
-	croak "Error $diag[0] while reading file $file: $diag[1]";
+	my @diag = _csv_diag ($csv);
+	$diag[0] == 2012 and return; # Also EOF (broken in Text::CSV_XS-1.10)
+
+	my $file = $tbl->{f_fqfn};
+	croak "Error $diag[0] while reading file $file: $diag[1] \@ line $diag[3] pos $diag[2]";
 	}
     @$fields < @{$tbl->{col_names}} and
 	push @$fields, (undef) x (@{$tbl->{col_names}} - @$fields);
     $self->{row} = (@$fields ? $fields : undef);
     } # fetch_row
 
-sub push_row
-{
+sub push_row {
     my ($self, $data, $fields) = @_;
-    my $tbl = $DBD::File::VERSION <= 0.38 ? $self : $self->{meta};
+    my $tbl = $self->{meta};
     my $csv = $self->{csv_csv_out};
     my $fh  = $tbl->{fh};
 
     unless ($csv->print ($fh, $fields)) {
-	my @diag = $csv->error_diag;
-	my $file = $DBD::File::VERSION <= 0.38 ? $self->{file} : $tbl->{f_fqfn};
-	croak "Error $diag[0] while writing file $file: $diag[1]";
+	my @diag = _csv_diag ($csv);
+	my $file = $tbl->{f_fqfn};
+	return do { $data->set_err ($DBI::stderr,
+	    "Error $diag[0] while writing file $file: $diag[1] \@ line $diag[3] pos $diag[2]"); undef };
 	}
     1;
     } # push_row
+
+no warnings 'once';
 *push_names = \&push_row;
+use warnings;
 
 1;
 
@@ -506,23 +400,24 @@ DBD::CSV - DBI driver for CSV files
 
     use DBI;
     # See "Creating database handle" below
-    $dbh = DBI->connect ("dbi:CSV:") or
-	die "Cannot connect: $DBI::errstr";
+    $dbh = DBI->connect ("dbi:CSV:", undef, undef, {
+        f_ext      => ".csv/r",
+        RaiseError => 1,
+        }) or die "Cannot connect: $DBI::errstr";
 
     # Simple statements
-    $dbh->do ("CREATE TABLE a (id INTEGER, name CHAR (10))") or
-	die "Cannot prepare: " . $dbh->errstr ();
+    $dbh->do ("CREATE TABLE foo (id INTEGER, name CHAR (10))");
 
     # Selecting
-    $dbh->{RaiseError} = 1;
     my $sth = $dbh->prepare ("select * from foo");
     $sth->execute;
-    while (my @row = $sth->fetchrow_array) {
-	print "id: $row[0], name: $row[1]\n";
+    $sth->bind_columns (\my ($id, $name));
+    while ($sth->fetch) {
+	print "id: $id, name: $name\n";
 	}
 
     # Updates
-    my $sth = $dbh->prepare ("UPDATE a SET name = ? WHERE id = ?");
+    my $sth = $dbh->prepare ("UPDATE foo SET name = ? WHERE id = ?");
     $sth->execute ("DBI rocks!", 1);
     $sth->finish;
 
@@ -556,17 +451,21 @@ from any CPAN mirror, for example
 =over 4
 
 =item DBI
+X<DBI>
 
-The DBI (Database independent interface for Perl), version 1.00 or
-a later release
+A recent version of the L<DBI> (Database independent interface for Perl).
+See below why.
 
 =item DBD::File
+X<DBD::File>
 
 This is the base class for DBD::CSV, and it is part of the DBI
-distribution. As DBD::CSV requires version 0.38 or newer for DBD::File
-it effectively requires DBI version 1.611 or newer.
+distribution. As DBD::CSV requires a matching version of L<DBD::File>
+which is (partly) developed by the same team that maintains
+DBD::CSV. See META.json or Makefile.PL for the minimum versions.
 
 =item SQL::Statement
+X<SQL::Statement>
 
 A simple SQL engine. This module defines all of the SQL syntax for
 DBD::CSV, new SQL support is added with each release so you should
@@ -575,11 +474,13 @@ look for updates to SQL::Statement regularly.
 It is possible to run C<DBD::CSV> without this module if you define
 the environment variable C<$DBI_SQL_NANO> to 1. This will reduce the
 SQL support a lot though. See L<DBI::SQL::Nano> for more details. Note
-that the test suite does not test in this mode!
+that the test suite does only test in this mode in the development
+environment.
 
 =item Text::CSV_XS
+X<Text::CSV_XS>
 
-This module is used for writing rows to or reading rows from CSV files.
+This module is used to read and write rows in a CSV file.
 
 =back
 
@@ -588,7 +489,7 @@ This module is used for writing rows to or reading rows from CSV files.
 Installing this module (and the prerequisites from above) is quite simple.
 The simplest way is to install the bundle:
 
-    $ cpan Bundle::CSV
+    $ cpan Bundle::DBD::CSV
 
 Alternatively, you can name them all
 
@@ -601,12 +502,12 @@ or even trust C<cpan> to resolve all dependencies for you:
 If you cannot, for whatever reason, use cpan, fetch all modules from
 CPAN, and build with a sequence like:
 
-    gzip -d < DBD-CSV-0.28.tgz | tar xf -
+    gzip -d < DBD-CSV-0.40.tgz | tar xf -
 
 (this is for Unix users, Windows users would prefer WinZip or something
 similar) and then enter the following:
 
-    cd DBD-CSV-0.28
+    cd DBD-CSV-0.40
     perl Makefile.PL
     make test
 
@@ -620,12 +521,14 @@ on installing in your own directories. L<ExtUtils::MakeMaker>.
 
 =head2 Supported SQL Syntax
 
-All SQL processing for DBD::CSV is done by the L<SQL::Statement> module.
+All SQL processing for DBD::CSV is done by SQL::Statement. See
+L<SQL::Statement> for more specific information about its feature set.
 Features include joins, aliases, built-in and user-defined functions,
 and more.  See L<SQL::Statement::Syntax> for a description of the SQL
 syntax supported in DBD::CSV.
 
-Table names are case insensitive unless quoted.
+Table- and column-names are case insensitive unless quoted. Column names
+will be sanitized unless L</raw_header> is true.
 
 =head1 Using DBD::CSV with DBI
 
@@ -657,26 +560,36 @@ The preferred way of passing the arguments is by driver attributes:
 
     # specify most possible flags via driver flags
     $dbh = DBI->connect ("dbi:CSV:", undef, undef, {
-	f_schema         => undef,
-	f_dir            => "data",
-	f_ext            => ".csv/r",
-	f_lock           => 2,
-	f_encoding       => "utf8",
+        f_schema         => undef,
+        f_dir            => "data",
+        f_dir_search     => [],
+        f_ext            => ".csv/r",
+        f_lock           => 2,
+        f_encoding       => "utf8",
 
-	csv_eol          => "\r\n",
-	csv_sep_char     => ",",
-	csv_quote_char   => '"',
-	csv_escape_char  => '"',
-	csv_class        => "Text::CSV_XS",
-	csv_null         => 1,
-	csv_tables       => {
-	    info => { file => "info.csv" }
-	    },
+        csv_eol          => "\r\n",
+        csv_sep_char     => ",",
+        csv_quote_char   => '"',
+        csv_escape_char  => '"',
+        csv_class        => "Text::CSV_XS",
+        csv_null         => 1,
+        csv_bom          => 0,
+        csv_tables       => {
+            syspwd => {
+                sep_char    => ":",
+                quote_char  => undef,
+                escape_char => undef,
+                file        => "/etc/passwd",
+                col_names   => [qw( login password
+                                    uid gid realname
+                                    directory shell )],
+		},
+            },
 
-	RaiseError       => 1,
-	PrintError       => 1,
-	FetchHashKeyName => "NAME_lc",
-	}) or die $DBI::errstr;
+        RaiseError       => 1,
+        PrintError       => 1,
+        FetchHashKeyName => "NAME_lc",
+        }) or die $DBI::errstr;
 
 but you may set these attributes in the DSN as well, separated by semicolons.
 Pay attention to the semi-colon for C<csv_sep_char> (as seen in many CSV
@@ -691,9 +604,16 @@ otherwise be seen as attribute separator:
 
 Using attributes in the DSN is easier to use when the DSN is derived from an
 outside source (environment variable, database entry, or configure file),
-whereas using all entries in the attribute hash is easier to read and to
+whereas specifying entries in the attribute hash is easier to read and to
 maintain.
 
+The default value for C<csv_binary> is C<1> (True).
+
+The default value for C<csv_auto_diag> is <1>. Note that this might cause
+trouble on perl versions older than 5.8.9, so up to and including perl
+version 5.8.8 it might be required to use C<;csv_auto_diag=0> inside the
+C<DSN> or C<csv_auto_diag => 0> inside the attributes.
+
 =head2 Creating and dropping tables
 
 You can create and drop tables with commands like the following:
@@ -719,7 +639,7 @@ with "/", "./" or "../" and they must not contain white space.
 =head2 Inserting, fetching and modifying data
 
 The following examples insert some data in a table and fetch it back:
-First all data in the string:
+First, an example where the column data is concatenated in the SQL string:
 
     $dbh->do ("INSERT INTO $table VALUES (1, ".
 	       $dbh->quote ("foobar") . ")");
@@ -727,18 +647,18 @@ First all data in the string:
 Note the use of the quote method for escaping the word "foobar". Any
 string must be escaped, even if it does not contain binary data.
 
-Next an example using parameters:
+Next, an example using parameters:
 
     $dbh->do ("INSERT INTO $table VALUES (?, ?)", undef, 2,
 	      "It's a string!");
 
-Note that you don't need to use the quote method here, this is done
-automatically for you. This version is particularly well designed for
+Note that you don't need to quote column data passed as parameters.
+This version is particularly well designed for
 loops. Whenever performance is an issue, I recommend using this method.
 
 You might wonder about the C<undef>. Don't wonder, just take it as it
-is. :-) It's an attribute argument that I have never ever used and
-will be parsed to the prepare method as a second argument.
+is. :-) It's an attribute argument that I have never used and will be
+passed to the prepare method as the second argument.
 
 To retrieve data, you can use the following:
 
@@ -790,8 +710,9 @@ Likewise you use the DELETE statement for removing rows:
 
 =head2 Error handling
 
-In the above examples we have never cared about return codes. Of course,
-this cannot be recommended. Instead we should have written (for example):
+In the above examples we have never cared about return codes. Of
+course, this is not recommended. Instead we should have written (for
+example):
 
     my $sth = $dbh->prepare ("SELECT * FROM $table WHERE id = ?") or
 	die "prepare: " . $dbh->errstr ();
@@ -847,30 +768,38 @@ The following DBI attributes are handled by DBD::File:
 =over 4
 
 =item AutoCommit
+X<AutoCommit>
 
 Always on
 
 =item ChopBlanks
+X<ChopBlanks>
 
 Works
 
 =item NUM_OF_FIELDS
+X<NUM_OF_FIELDS>
 
 Valid after C<$sth-E<gt>execute>
 
 =item NUM_OF_PARAMS
+X<NUM_OF_PARAMS>
 
 Valid after C<$sth-E<gt>prepare>
 
 =item NAME
+X<NAME>
 
 =item NAME_lc
+X<NAME_lc>
 
 =item NAME_uc
+X<NAME_uc>
 
 Valid after C<$sth-E<gt>execute>; undef for Non-Select statements.
 
 =item NULLABLE
+X<NULLABLE>
 
 Not really working. Always returns an array ref of one's, as DBD::CSV
 does not verify input data. Valid after C<$sth-E<gt>execute>; undef for
@@ -890,19 +819,37 @@ These attributes and methods are not supported:
 In addition to the DBI attributes, you can use the following dbh
 attributes:
 
+=head2 DBD::File attributes
+
 =over 4
 
 =item f_dir
+X<f_dir>
 
 This attribute is used for setting the directory where CSV files are
-opened. Usually you set it in the dbh, it defaults to the current
-directory ("."). However, it is overwritable in the statement handles.
+opened. Usually you set it in the dbh and it defaults to the current
+directory ("."). However, it may be overridden in statement handles.
+
+=item f_dir_search
+X<f_dir_search>
+
+This attribute optionally defines a list of extra directories to search
+when opening existing tables. It should be an anonymous list or an array
+reference listing all folders where tables could be found.
+
+    my $dbh = DBI->connect ("dbi:CSV:", "", "", {
+	f_dir        => "data",
+	f_dir_search => [ "ref/data", "ref/old" ],
+	f_ext        => ".csv/r",
+	}) or die $DBI::errstr;
 
 =item f_ext
+X<f_ext>
 
 This attribute is used for setting the file extension.
 
 =item f_schema
+X<f_schema>
 
 This attribute allows you to set the database schema name. The default is
 to use the owner of C<f_dir>. C<undef> is allowed, but not in the DSN part.
@@ -914,59 +861,88 @@ to use the owner of C<f_dir>. C<undef> is allowed, but not in the DSN part.
 	}) or die $DBI::errstr;
 
 =item f_encoding
+X<f_encoding>
 
-This attribute allows you to set the encoding of the data. With CSV, it is
-not possible to set (and remember) the encoding on a per-field basis, but
-DBD::File now allows to set the encoding of the underlying file. If this
-attribute is not set, or undef is passed, the file will be seen as binary.
+This attribute allows you to set the encoding of the data. With CSV, it is not
+possible to set (and remember) the encoding on a column basis, but DBD::File
+now allows the encoding to be set on the underlying file. If this attribute is
+not set, or undef is passed, the file will be seen as binary.
 
 =item f_lock
+X<f_lock>
 
-With this attribute, you can force locking mode (if locking is supported
-at all) for opening tables. By default, tables are opened with a shared
-lock for reading, and with an exclusive lock for writing. The supported
-modes are:
+With this attribute you can specify a locking mode to be used (if locking is
+supported at all) for opening tables. By default, tables are opened with a
+shared lock for reading, and with an exclusive lock for writing. The
+supported modes are:
 
 =over 2
 
 =item 0
+X<0>
 
 Force no locking at all.
 
 =item 1
+X<1>
 
 Only shared locks will be used.
 
 =item 2
+X<2>
 
 Only exclusive locks will be used.
 
 =back
 
+=back
+
 But see L<DBD::File/"KNOWN BUGS">.
 
-=item csv_eol
+=head2 DBD::CSV specific attributes
 
-=item csv_sep_char
-
-=item csv_quote_char
-
-=item csv_escape_char
+=over 4
 
 =item csv_class
 
+The attribute I<csv_class> controls the CSV parsing engine. This defaults
+to C<Text::CSV_XS>, but C<Text::CSV> can be used in some cases, too.
+Please be aware that C<Text::CSV> does not care about any edge case as
+C<Text::CSV_XS> does and that C<Text::CSV> is probably about 100 times
+slower than C<Text::CSV_XS>.
+
+=back
+
+=head2 Text::CSV_XS specific attributes
+
+=over 4
+
+=item csv_eol
+X<csv_eol>
+
+=item csv_sep_char
+X<csv_sep_char>
+
+=item csv_quote_char
+X<csv_quote_char>
+
+=item csv_escape_char
+X<csv_escape_char>
+
 =item csv_csv
+X<csv_csv>
 
 The attributes I<csv_eol>, I<csv_sep_char>, I<csv_quote_char> and
 I<csv_escape_char> are corresponding to the respective attributes of the
-Text::CSV_XS object. You want to set these attributes if you have unusual
-CSV files like F</etc/passwd> or MS Excel generated CSV files with a semicolon
-as separator. Defaults are "\015\012", ';', '"' and '"', respectively.
+I<csv_class> (usually Text::CSV_CS) object. You may want to set these
+attributes if you have unusual CSV files like F</etc/passwd> or MS Excel
+generated CSV files with a semicolon as separator. Defaults are
+C<\015\012>", C<,>, C<"> and C<">, respectively.
 
 The I<csv_eol> attribute defines the end-of-line pattern, which is better
 known as a record separator pattern since it separates records.  The default
-is windows-style end-of-lines "\015\012" for output (writing) and unset for
-input (reading), so if on unix you may want to set this to newline ("\n")
+is windows-style end-of-lines C<\015\012> for output (writing) and unset for
+input (reading), so if on unix you may want to set this to newline (C<\n>)
 like this:
 
   $dbh->{csv_eol} = "\n";
@@ -999,6 +975,7 @@ Additionally you may overwrite these attributes on a per-table base in
 the I<csv_tables> attribute.
 
 =item csv_null
+X<csv_null>
 
 With this option set, all new statement handles will set C<always_quote>
 and C<blank_is_undef> in the CSV parser and writer, so it knows how to
@@ -1009,17 +986,45 @@ reset it with a false value. You can pass it to connect, or set it later:
 
   $dbh->{csv_null} = 1;
 
+=item csv_bom
+X<csv_bom>
+
+With this option set, the CSV parser will try to detect BOM (Byte Order Mark)
+in the header line. This requires L<Text::CSV_XS> version 1.22 or higher.
+
+  $dbh = DBI->connect ("dbi:CSV:", "", "", { csv_bom => 1 });
+
+  $dbh->{csv_bom} = 1;
+
 =item csv_tables
+X<csv_tables>
 
 This hash ref is used for storing table dependent metadata. For any
 table it contains an element with the table name as key and another
 hash ref with the following attributes:
 
+=over 4
+
+=item o
+
+All valid attributes to the CSV parsing module. Any of the can optionally
+be prefixed with C<csv_>.
+
+=item o
+
+All attributes valid to DBD::File
+
+=back
+
+If you pass it C<f_file> or its alias C<file>, C<f_ext> has no effect, but
+C<f_dir> and C<f_encoding> still have.
+
 =item csv_*
+X<csv_*>
 
 All other attributes that start with C<csv_> and are not described above
-will be passed to C<Text::CSV_XS> (without the C<csv_> prefix). these
-extra options are most likely to be only useful for reading (select)
+will be passed to C<Text::CSV_XS> (without the C<csv_> prefix). These
+extra options are only likely to be useful for reading (select)
 handles. Examples:
 
   $dbh->{csv_allow_whitespace}    = 1;
@@ -1028,33 +1033,46 @@ handles. Examples:
 
 See the C<Text::CSV_XS> documentation for the full list and the documentation.
 
+=back
+
+=head2 Driver specific attributes
+
 =over 4
 
-=item file
+=item f_file
+X<f_file>
 
-The tables file name; defaults to
+The name of the file used for the table; defaults to
 
     "$dbh->{f_dir}/$table"
 
 =item eol
+X<eol>
 
 =item sep_char
+X<sep_char>
 
 =item quote_char
+X<quote_char>
 
 =item escape_char
+X<escape_char>
 
 =item class
+X<class>
 
 =item csv
+X<csv>
 
 These correspond to the attributes I<csv_eol>, I<csv_sep_char>,
 I<csv_quote_char>, I<csv_escape_char>, I<csv_class> and I<csv_csv>.
-The difference is that they work on a per-table base.
+The difference is that they work on a per-table basis.
 
 =item col_names
+X<col_names>
 
 =item skip_first_row
+X<skip_first_row>
 
 By default DBD::CSV assumes that column names are stored in the first row
 of the CSV file and sanitizes them (see C<raw_header> below). If this is
@@ -1066,24 +1084,57 @@ If you supply an empty array ref, the driver will read the first row
 for you, count the number of columns and create column names like
 C<col0>, C<col1>, ...
 
+Note that column names that match reserved SQL words will cause unwanted
+and sometimes confusing errors. If your CSV has headers that match reserved
+words, you will require these two attributes.
+
+If C<test.csv> looks like
+
+ select,from
+ 1,2
+
+the select query would result in C<select select, from from test;>, which
+obviously is illegal SQL.
+
 =item raw_header
+X<raw_header>
 
 Due to the SQL standard, field names cannot contain special characters
-like a dot (C<.>). Following the approach of mdb_tools, all these tokens
-are translated to an underscore (C<_>) when reading the first line of the
-CSV file, so all field names are `sanitized'. If you do not want this to
-happen, set C<raw_header> to a true value. DBD::CSV cannot guarantee that
-any part in the toolchain will work if field names have those characters,
+like a dot (C<.>) or a space (C< >) unless the column names are quoted.
+Following the approach of mdb_tools, all these tokens are translated to an
+underscore (C<_>) when reading the first line of the CSV file, so all field
+names are 'sanitized'. If you do not want this to happen, set C<raw_header>
+to a true value and the entries in the first line of the CSV data will be
+used verbatim for column headers and field names.  DBD::CSV cannot guarantee
+that any part in the toolchain will work if field names have those characters,
 and the chances are high that the SQL statements will fail.
 
-=back
+Currently, the sanitizing of headers is as simple as
+
+  s/\W/_/g;
+
+Note that headers (column names) might be folded in other parts of the code
+stack, specifically SQL::Statement, whose docs mention:
+
+ Wildcards are expanded to lower cased identifiers. This might
+ confuse some people, but it was easier to implement.
+
+That means that in
+
+ my $sth = $dbh->prepare ("select * from foo");
+ $sth->execute;
+ while (my $row = $sth->fetchrow_hashref) {
+     say for keys %$row;
+     }
+
+all keys will show as all lower case, regardless of the original header.
 
 =back
 
 It's strongly recommended to check the attributes supported by
 L<DBD::File/Metadata>.
 
-Example: Suggest you want to use F</etc/passwd> as a CSV file. :-)
+Example: Suppose you want to use F</etc/passwd> as a CSV file. :-)
 There simplest way is:
 
     use DBI;
@@ -1100,7 +1151,7 @@ There simplest way is:
     $sth = $dbh->prepare ("SELECT * FROM passwd");
 
 Another possibility where you leave all the defaults as they are and
-overwrite them on a per table base:
+override them on a per table basis:
 
     require DBI;
     my $dbh = DBI->connect ("dbi:CSV:");
@@ -1109,7 +1160,7 @@ overwrite them on a per table base:
 	sep_char    => ":",
 	quote_char  => undef,
 	escape_char => undef,
-	file        => "/etc/passwd",
+	f_file      => "/etc/passwd",
 	col_names   => [qw( login password uid gid
 			    realname directory shell )],
 	};
@@ -1122,6 +1173,7 @@ These methods are inherited from DBD::File:
 =over 4
 
 =item data_sources
+X<data_sources>
 
 The C<data_sources> method returns a list of sub-directories of the current
 directory in the form "dbi:CSV:directory=$dirname".
@@ -1132,8 +1184,9 @@ If you want to read the sub-directories of another directory, use
     my @list = $drh->data_sources (f_dir => "/usr/local/csv_data");
 
 =item list_tables
+X<list_tables>
 
-This method returns a list of file names inside $dbh->{directory}.
+This method returns a list of file-names inside $dbh->{directory}.
 Example:
 
     my $dbh  = DBI->connect ("dbi:CSV:directory=/usr/local/csv_data");
@@ -1152,9 +1205,9 @@ L<Creating and dropping tables> above.
 =item *
 
 The module is using flock () internally. However, this function is not
-available on platforms. Using flock () is disabled on MacOS and Windows
-95: There's no locking at all (perhaps not so important on these
-operating systems, as they are for single users anyways).
+available on some platforms. Use of flock () is disabled on MacOS and
+Windows 95: There's no locking at all (perhaps not so important on
+these operating systems, as they are for single users anyways).
 
 =back
 
@@ -1163,6 +1216,7 @@ operating systems, as they are for single users anyways).
 =over 4
 
 =item Tests
+X<Tests>
 
 Aim for a full 100% code coverage
 
@@ -1179,28 +1233,34 @@ Add tests that specifically test DBD::File functionality where
 that is useful.
 
 =item RT
+X<RT>
 
 Attack all open DBD::CSV bugs in RT
 
 =item CPAN::Forum
+X<CPAN::Forum>
 
 Attack all items in http://www.cpanforum.com/dist/DBD-CSV
 
 =item Documentation
+X<Documentation>
 
 Expand on error-handling, and document all possible errors.
 Use Text::CSV_XS::error_diag () wherever possible.
 
 =item Debugging
+X<Debugging>
 
 Implement and document dbd_verbose.
 
 =item Data dictionary
+X<Data dictionary>
 
 Investigate the possibility to store the data dictionary in a file like
 .sys$columns that can store the field attributes (type, key, nullable).
 
 =item Examples
+X<Examples>
 
 Make more real-life examples from the docs in examples/
 
@@ -1233,7 +1293,7 @@ Previous maintainer was Jeff Zucker
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2009-2010 by H.Merijn Brand
+Copyright (C) 2009-2018 by H.Merijn Brand
 Copyright (C) 2004-2009 by Jeff Zucker
 Copyright (C) 1998-2004 by Jochen Wiedmann
 
